// MeasurementTable.java

package astroj;

import ij.*;
import ij.measure.*;
import ij.text.*;

import java.awt.Frame;
import java.io.*;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;
import java.util.Vector;

/**
 * Like an ImageJ ResultsTable but with more functionality.
 *
 * @author F.V. Hessman, Georg-August-Universitaet Goettingen
 * @version 1.4
 * @date 2006-Sep-11
 */
public class MeasurementTable extends ResultsTable
	{
	public static String PREFIX = new String("Measurements");
	public static String RESULTS = new String("Results");
	protected String shortName = null;
	public static int DEFAULT_DECIMALS = 6;

	/**
	 * Creates an empty default MeasurementTable.
	 */
	public MeasurementTable ()
		{
		setPrecision(DEFAULT_DECIMALS);
		shortName = null;
		}

	/**
	 * Creates an empty MeasurementTable with a given long name.
	 */
	public MeasurementTable (String tableName)
		{
		setPrecision (DEFAULT_DECIMALS);
		shortName = MeasurementTable.shorterName (tableName);
		}

	/**
	 * Creates a MeasurementTable object from a TextWindow.
	 */
	public MeasurementTable (TextWindow w)
		{
		super();
		setPrecision (DEFAULT_DECIMALS);
		TextPanel panel = w.getTextPanel();
		shortName = MeasurementTable.shorterName (w.getTitle());
		int count = panel.getLineCount();
		// IJ.showMessage(panel.getText());
		}

	/**
	 * Transfers the last entry from a MeasurementTable to the standard Results table.
	 */
	public static void transferLastRow (String tableName)
		{
		MeasurementTable t = new MeasurementTable(tableName);
		ResultsTable r = ResultsTable.getResultsTable();
		if (t == null || r == null)
			{
			IJ.error("Unable to read measurement table or Results table!");
			return;
			}
		int row = t.getCounter();
		int ncols = t.getLastColumn();
IJ.log("row="+row+", ncols="+ncols);
		for (int col=0; col < ncols; col++)
			{
			String heading = t.getColumnHeading(col);
			double val = t.getValueAsDouble(col,row);
IJ.log(heading+" "+val);
			r.addValue(heading,val);
			}
		return;
		}
	
	/**
	 * Stores a number using a temporary different number of decimal places.
	 */
	public void addValue (String column, double value, int places)
		{
		setPrecision(places);
		super.addValue (column,value);
		// setPrecision (DEFAULT_DECIMALS);
		// PRESENT ResultsTable DOESN'T KEEP TRACK OF INDIVIDUAL PRECISIONS!!!
		}  

	/**
	 * Displays/Refreshes a MeasurementTable.
	 */
	public void show()
		{
		super.show (MeasurementTable.longerName(shortName));
		}

	/**
	 * ResultTable method to be overridden: a ResultTable's shortTitle is a MeasurementTable's longName.
	 */
	public String shortTitle()
		{
		return MeasurementTable.longerName(shortName);
		}

	/**
	 * Returns the ImagePlus from which the measurements in the MeasurementTable
	 * were obtained.  The image must still be displayed for the WindowManager to find it.
	 */
	public ImagePlus getImage()
		{
		if (shortName == null)
			return null;
		int ids[] = WindowManager.getIDList();
		int n = WindowManager.getWindowCount();
		for (int i=0; i < n; i++)
			{
			ImagePlus img = WindowManager.getImage(ids[i]);
			if (shortName.equals(img.getTitle()))	// getShortTitle()?
				return img;
			}
		return null;
		}

	/**
	 * Class method for extracting a TextPanel from a MeasurementTable with a given name.
	 */
	public static TextPanel getTextPanel (String tableName)
		{
		Frame frame = WindowManager.getFrame(tableName);
		if (frame == null || !(frame instanceof TextWindow))
			{
			// IJ.error ("Unable to access "+tableName+"!");
			return null;
			}
		else	{
			TextWindow win = (TextWindow)frame;
			return win.getTextPanel();
			}
		}

	/**
	 * Indicates whether a MeasurementTable exists with the given name.
	 */
	public static boolean exists (String tableName)
		{
		TextPanel tp = getTextPanel (tableName);
		if (tp == null)
			return false;
		else
			return true;
		}

	/**
	 * Returns a MeasurementTable reconstructed from a text file produced by ImageJ from a MeasurementTable/ResultsTable.
	 */
	public static MeasurementTable getTableFromFile (String filename)
		{
		BufferedReader in = null;
		MeasurementTable table = null;

		try	{
			in = new BufferedReader(new FileReader(filename));

			// READ HEADER LINE

			String line = in.readLine();
			if (line == null)
				{
				IJ.error("MeasurementTable: cannot read header line!");
				in.close();
				return null;
				}
// IJ.showMessage("header:"+line);

			// OPEN NEW MeasurementTable

			table = new MeasurementTable();	// SHOULD USE filename); BUT CAN'T BECAUSE OF PRIVATE WindowManager.getNonImageWindows() METHOD

			// CREATE COLUMNS

			String[] labels = line.split("\t");
			int n = labels.length;
// String nix = "";
////			for (int l=0; l < n; l++)
////				{
////				table.setHeading (l,labels[l]);
// nix += labels[l]+",";
////				}
// IJ.showMessage("labels: "+nix);

			// READ DATA

			int row = 0;
			double d = 0.0;
			boolean usesText = false;
			while ((line = in.readLine()) != null)
				{
// IJ.showMessage(""+row+": "+line);
				table.incrementCounter();

				String[] words = line.trim().split("\t");
				int m = words.length;
// if (m != n) IJ.showMessage("n="+n+", m="+m);

				if (row == 0)
					{
					try	{
						d = Double.parseDouble(words[0]);
						d = Double.parseDouble(words[1]);
						}
					catch (NumberFormatException e)
						{
						usesText = true;
// IJ.showMessage("usesText : "+words[0]+" "+words[1]);
						}
					}
				for (int col=1; col < m; col++)
					{
// IJ.showMessage("words["+col+"]="+words[col]);
					if (col == 1 && usesText)
						table.addLabel(labels[1],words[1]);
					else	{
						if (words[col] == null)
							d = 0.0;	// Double.NaN;
						else if (words[col].trim().equals("") || words[col].trim().equals("-"))
							d = 0.0;	// Double.NaN;
						else if (isHMS(words[col]))
							d = hms(words[col]);
						else
							d = Double.parseDouble(words[col]);
                        table.addValue(labels[col],d);
////						table.setValue(col,row,d);
						}
					}
				row++;
				}
			in.close();
			}
		catch (IOException e)
			{
			System.err.println ("MeasurementTable: "+e.getMessage());
			IJ.error("MeasurementTable: "+e.getMessage());
			table = null;
			}
		catch (NumberFormatException nfe)
			{
			System.err.println ("MeasurementTable: "+nfe.getMessage());
			IJ.error("MeasurementTable: "+nfe.getMessage());
			table = null;
			}
		try { in.close(); } catch(Exception exc) {}
		return table;
		}

	/**
	 * Returns an existing MeasurementTable reconstructed from the TextWindow with the appropriate name.
	 */
	public static MeasurementTable getTable (String tableName)
		{
		int n;
		String str="";

		String name = MeasurementTable.longerName (tableName);

		// CREATE EMPTY TABLE

		MeasurementTable table = new MeasurementTable (name);

		// IF NO PREVIOUS TABLE EXISTS, RETURN EMPTY TABLE

		if (! exists(name))
			return table;

		// GET CONTENTS OF EXISTING MeasurementTable

		TextPanel panel = getTextPanel (name);

        String[] headings = panel.getColumnHeadings().split("\t");
        StringTokenizer parser = new StringTokenizer ("","\t");

		n = panel.getLineCount();
		for (int row=0; row < n; row++)
			{

			parser = new StringTokenizer (panel.getLine(row),"\t");
			str = parser.nextToken();		// IGNORE MEASUREMENT NUMBER

			table.incrementCounter();
            for(int column = 1; column < headings.length; column++)
                {
                if (parser.hasMoreTokens() && (str = parser.nextToken()) != null)
                    {
                    try	{
                        double d = Double.parseDouble(str);
                        table.addValue(headings[column],d);
                        }
                    catch (NumberFormatException e)
                        {
                        table.addLabel(headings[1],str);
                        }
                    }
                else
                    {
                    table.addValue(headings[column],0.0);
                    }
                }
			}

		return table;
		}

	/**
	 * Returns a MeasurementTable with the path pathname.  If already exists, the table in reconstructed.
	 */
	public static MeasurementTable readTable (String pathname)
		{
		// FILTER pathname HERE ONE OF THESE DAYS!!!!

		// CREATE EMPTY TABLE

		MeasurementTable table = new MeasurementTable (pathname);

		try	{
			int column=-1;
			int row = -1;
			FileInputStream stream = new FileInputStream(pathname);
			BufferedReader reader = new BufferedReader(new InputStreamReader(stream));

			StreamTokenizer parser = new StreamTokenizer(reader);
			parser.eolIsSignificant(true);
			parser.whitespaceChars((int)'\t',(int)'\t') ;     // JUST HTABS
			parser.wordChars((int)'!',(int)'~');

			int ttype=0;
			while (parser.nextToken() != StreamTokenizer.TT_EOF)
				{
				ttype = parser.ttype;
				if (ttype == StreamTokenizer.TT_EOL)
					{
					column=-1;
					row++;
					}
				else if (row == -1)
					{
					if ( ttype == StreamTokenizer.TT_WORD)
						table.setHeading(++column, parser.sval);
					// IJ.showMessage("row="+row+", column="+column+", label="+parser.sval);
					}
				else if (ttype == StreamTokenizer.TT_NUMBER)
					{
					if (column == -1)
						table.incrementCounter();
					else
						table.setValue(column,row,parser.nval);
					column++;
					}
				}
			reader.close();
			stream.close();
			}
		catch (FileNotFoundException e)
			{
			IJ.error("Error: "+e.getMessage());
			}
		catch (IOException e)
			{
			IJ.error("Error: "+e.getMessage());
			}
		return table;
		}

	/**
	 * Retrieves a double array column.
	 */
	public double[] getDoubleColumn (int col)
		{
		int cntr = this.getCounter();
		double[] d = new double[cntr];
		for (int r=0; r < cntr; r++)
			{
			double dd = this.getValueAsDouble(col,r);
// IJ.log("arr["+col+"]["+r+"]="+dd);
			d[r] = dd;
			}
/* OLD VERSION DOESN'T WORK IF TABLE CONTAINS NaN!!!
		float[] f = this.getColumn(col);
		if (f == null) return null;

		double[] d = new double[f.length];
		for (int i=0; i < f.length; i++)
			d[i] = f[i];
*/
		return d;
		}

	/**
	 * Inserts a double array column.
	 */
	public boolean putDoubleColumn (String title, double[] arr)
		{
		int col = getFreeColumn (title);
		if (col == COLUMN_IN_USE || col == TABLE_FULL)
			return false;

		for (int i=0; i < arr.length; i++)
			setValue (col, i, arr[i]);
		return true;
		}

	/**
	 * Returns the full name of a MeasurementTable, including the standard prefix if not already given.
	 */
	public static String longerName (String name)
		{
		if (name == null)
			return new String (MeasurementTable.PREFIX);
		else if (name.equals(MeasurementTable.RESULTS))
			return name;
		else if (name.equals(MeasurementTable.PREFIX))
			return name;
		else if (name.startsWith(MeasurementTable.PREFIX+" in "))
			return name;
		else if (name.indexOf(".txt") >= 0)
			return name;
		else
			return new String (MeasurementTable.PREFIX+" in "+name);
		}

	/**
	 * Returns the part of the MeasurementTable name without the standard prefix.
	 */
	public static String shorterName (String tableName)
		{
		if (tableName == null)
			return null;
		else if (tableName.equals (MeasurementTable.PREFIX))
			return null;
		else if (tableName.equals (MeasurementTable.RESULTS))
			return null;
		else if (tableName.startsWith(MeasurementTable.PREFIX+" in "))
			return tableName.substring(tableName.indexOf(" in ")+4);
		else
			return tableName;
		}

	/**
	 * This method desperately attempts to replace the private functionality of WindowManager.getNonImageWindows() (before ImageJ Version 1.38q)
	 */
	public static String[] getMeasurementTableNames()
		{
		Vector frames = new Vector();

		// FIND MEASUREMENT TABLES ASSOCIATED WITH IMAGES

		if (IJ.versionLessThan("1.40"))
			{
			Frame std = WindowManager.getFrame ("Results");
			if (std != null)
				frames.addElement ("Results");
			std = WindowManager.getFrame (PREFIX);
			if (std != null)
				frames.addElement (PREFIX);
			String[] openImages = IJU.listOfOpenImages (null);
			if (openImages != null && openImages.length > 0)
				{
				for (int i=0; i < openImages.length; i++)
					{
					String longName = longerName (openImages[i]);
					if (WindowManager.getFrame (longName) != null)
						frames.addElement (longName);
					}
				}
			}
		else	{
			Frame[] windows = WindowManager.getNonImageWindows();
			if (windows != null && windows.length > 0)
				{
				Frame std = WindowManager.getFrame ("Results");
				if (std != null)
					frames.addElement ("Results");
				for (int i=0; i < windows.length; i++)
					{
					String title = windows[i].getTitle();
					// if (title.startsWith("Measurement"))
						frames.addElement (title);
					}
				}
			}

		// ANYTHING AT ALL?

		int n = frames.size();
		if (n == 0) return null;

		// THEN GATHER INTO STRING ARRAY

		String[] result = new String [n];
		for (int i=0; i < n; i++)
			result[i] = new String ((String)frames.elementAt (i));
		return result;
		}

	/**
	 * Checks to see if a string contains a hh:mm:ss.sss representation of an angle/time.
	 */
	static boolean isHMS (String s)
		{
		String[] arr = s.split(":");
		if (arr.length > 1)
			return true;
		else
			return false;
		}

	/**
	 * Converts hh:mm:ss.sss to a number.
	 */
	static double hms (String s) 
		{
		double d[] = new double[] {0.0,0.0,0.0};
		String[] arr = s.split(":");
		int n = (arr.length > 3)? 3 : arr.length;
		double sgn = 1.0;
		for (int i=0; i < n; i++)
			{
			d[i] = Double.parseDouble(arr[i]);
			if (arr[i].trim().startsWith("-")) sgn = -1.0;
			}
		return sgn*(Math.abs(d[0])+Math.abs(d[1])/60.0+Math.abs(d[2])/3600.0);
		}
	}
